QWidget {
    background-color: #4c6085;
}

QPushButton {
  background-clip: padding-box;
  background-color: #13c4a3;
  border: 1px solid transparent;
  border-radius: .25rem;
  color: #000;
  font-family: system-ui,-apple-system,system-ui,"Helvetica Neue",Helvetica,Arial,sans-serif;
  font-size: 16px;
  font-weight: 600;
  line-height: 1.25;
  margin: 0;
  min-height: 3rem;
  padding: calc(.875rem - 1px) calc(1.5rem - 1px);
  position: relative;
  text-decoration: none;
  vertical-align: baseline;
  width: auto;
}

QPushButton:hover{
  background-color: #36f1cd;
}

QPushButton:disabled{
  background-color: darkgray;
}

QComboBox {
    background-color: #13c4a3;
    font-family: system-ui,-apple-system,system-ui,"Helvetica Neue",Helvetica,Arial,sans-serif;
    font-size: 16px;
    font-weight: 600;
}

QComboBox:hover{
    background-color: #36f1cd;
}

QComboBox:disabled{
  background-color: darkgray;
}

QListWidget {
  background-color: #13c4a3;
  font-family: system-ui,-apple-system,system-ui,"Helvetica Neue",Helvetica,Arial,sans-serif;
  font-size: 16px;
  font-weight: 600;
}

QListWidget:hover{
  background-color: #36f1cd;
}


QRadioButton {
  background-color: #36f1cd;
}

QLabel {
  font-size: 12pt;
}

QCheckBox::indicator:checked {
  background-color: #36f1cd;
}
