__all__: list = []
__author__ = "Diamond Light Source - Scientific Software"
__email__ = "scientificsoftware@diamond.ac.uk"
__version__ = "0.1.0"
__version_tuple__ = tuple(int(x) for x in __version__.split("."))
